import subprocess
from typing import Any, List, Optional, Tuple

from simple_logger.logger import get_logger

from pyhelper_utils.exceptions import CommandExecFailed
from rrmngmnt import Host

LOGGER = get_logger(name=__name__)

TIMEOUT_30MIN = 30 * 60


def run_command(
    command: List[str],
    verify_stderr: bool = True,
    shell: bool = False,
    timeout: Optional[int] = None,
    capture_output: bool = True,
    check: bool = True,
    hide_log_command: bool = False,
    **kwargs: Any,
) -> Tuple[bool, str, str]:
    """
    Run command locally.

    Args:
        command (list): Command to run
        verify_stderr (bool, default True): Check command stderr
        shell (bool, default False): run subprocess with shell toggle
        timeout (int, optional): Command wait timeout
        capture_output (bool, default False): Capture command output
        check (bool, default True):  If check is True and the exit code was non-zero, it raises a
            CalledProcessError
        hide_log_command (bool, default False): If hide_log_command is True and check will be set to False,
            CalledProcessError will not get raise and command will not be printed.

    Returns:
        tuple: True, out if command succeeded, False, err otherwise.

    Raises:
        CalledProcessError: when check is True and command execution fails
    """
    command_for_log = ["Hide", "By", "User"] if hide_log_command else command

    LOGGER.info(f"Running {' '.join(command_for_log)} command")

    # when hide_log_command is set to True, check should be set to False to avoid logging sensitive data in
    # the exception
    sub_process = subprocess.run(
        command,
        capture_output=capture_output,
        check=check if not hide_log_command else False,
        shell=shell,
        text=True,
        timeout=timeout,
        **kwargs,
    )
    out_decoded = sub_process.stdout
    err_decoded = sub_process.stderr

    error_msg = (
        f"Failed to run {command_for_log}. rc: {sub_process.returncode}, out: {out_decoded}, error: {err_decoded}"
    )

    if sub_process.returncode != 0:
        LOGGER.error(error_msg)
        return False, out_decoded, err_decoded

    # From this point and onwards we are guaranteed that sub_process.returncode == 0
    if err_decoded and verify_stderr:
        LOGGER.error(error_msg)
        return False, out_decoded, err_decoded

    return True, out_decoded, err_decoded


def run_ssh_commands(
    host: Host,
    commands: List[str],
    get_pty: bool = False,
    check_rc: bool = True,
    timeout: int = TIMEOUT_30MIN,
    tcp_timeout: Optional[float] = None,
) -> list:
    """
    Run commands on remote host via SSH

    Args:
        host (Host): rrmngmnt host to execute the commands from.
        commands (list): List of multiple command lists [[cmd1, cmd2, cmd3]] or a list with a single command [cmd]
            Examples:
                 single command: shlex.split("sudo reboot"),
                 multiple commands: [shlex.split("sleep 5"), shlex.split("date")]

        get_pty (bool): get_pty parameter for remote session (equivalent to -t argument for ssh)
        check_rc (bool): if True checks command return code and raises if rc != 0
        timeout (int): ssh exec timeout
        tcp_timeout (float): an optional timeout (in seconds) for the TCP connect

    Returns:
        list: List of commands output.

    Raise:
        CommandExecFailed: If command failed to execute.
    """
    results: List[str] = []
    commands_list: List[List[str]] = commands if isinstance(commands[0], list) else [commands]
    with host.executor().session(timeout=tcp_timeout) as ssh_session:
        for cmd in commands_list:
            rc, out, err = ssh_session.run_cmd(cmd=cmd, get_pty=get_pty, timeout=timeout)
            LOGGER.info(f"[SSH][{host.fqdn}] Executed: {' '.join(cmd)}, rc:{rc}, out: {out}, error: {err}")
            if rc and check_rc:
                raise CommandExecFailed(name=" ".join(cmd), err=err)

            results.append(out)

    return results



import time
import rich
import datetime
import sys
from typing import Callable, Any
from simple_logger.logger import get_logger

LOGGER = get_logger(name="runners")


def function_runner_with_pdb(func: Callable, dry_run: bool = False, *args: Any, **kwargs: Any) -> Any:
    """
    Run function with support to drop into pdb.

    In order to use this function you need to have --pdb in sys.argv

    Example:
        @click.option(
            "--pdb",
            help="Drop to `ipdb` shell on exception",
            is_flag=True,
            show_default=True,
            )
        def main(pdb):
            <code>

        if __name__ == "__main__":
            function_runner_with_pdb(func=main)

    Args:
        func (Callable): Function to run
        dry_run (bool, optional): Run without drop into pdb. Defaults to False.

    Returns:
        None: if execution was successful.

    Raises:
        SystemExit: If --pdb is not in sys.argv and execution failed
    """
    start_time = time.time()
    should_raise = False

    try:
        return func(*args, **kwargs)
    except Exception as ex:
        if "--pdb" in sys.argv:
            _, _, tb = sys.exc_info()
            if not dry_run:
                ipdb = __import__("ipdb")  # Bypass debug-statements pre-commit hook
                ipdb.post_mortem(tb)
        else:
            rich.print(f"Failed to execute with Error {ex}")
            should_raise = True
    finally:
        total_time = datetime.timedelta(seconds=time.time() - start_time)
        rich.print(f"Total execution time: {total_time}")
        if should_raise:
            sys.exit(1)



from logging import Logger
from typing import Optional
import requests
import json


def send_slack_message(
    message: str,
    webhook_url: Optional[str] = None,
    logger: Optional[Logger] = None,
    post_timout: int = 30,
    raise_on_error: bool = True,
) -> None:
    """
    Send slack message via slack webhook url.

    Args:
        message (str): message to send.
        webhook_url (str): slack webhook url.
        logger (Logger): logger to use, if not passed no logs will be displayed.
        post_timout (int): Timeout in seconds to send to request.post.
        raise_on_error (bool): Raise exception if error.
    """
    try:
        if webhook_url:
            slack_data = {"text": message}
            if logger:
                logger.info(f"Sending message to slack: {message}")

            response = requests.post(
                webhook_url,
                data=json.dumps(slack_data),
                headers={"Content-Type": "application/json"},
                timeout=post_timout,
            )
            if response.status_code != 200:
                if logger:
                    logger.error(
                        f"Request to slack returned an error {response.status_code} with the following message: {response.text}"
                    )
    except Exception as ex:
        if logger:
            logger.error(f"Failed to send slack message. error: {ex}")

        if raise_on_error:
            raise



from __future__ import annotations
import datetime
import re
from time import sleep
from functools import wraps
from logging import Logger
from typing import Any, Callable


def tts(ts: Any) -> int:
    """
    Convert time string to seconds.

    Args:
        ts (str): time string to convert, can be and int followed by s/m/h
            if only numbers was sent return int(ts)

    Example:
        >>> tts(ts="1h")
        3600
        >>> tts(ts="3600")
        3600

    Returns:
        int: Time in seconds
    """
    if time_and_unit_match := re.match(r"(?P<time>\d+)(?P<unit>\w)", str(ts)):
        time_and_unit = time_and_unit_match.groupdict()
    else:
        return int(ts)

    _time = int(time_and_unit["time"])
    _unit = time_and_unit["unit"].lower()
    if _unit == "s":
        return _time
    elif _unit == "m":
        return _time * 60
    elif _unit == "h":
        return _time * 60 * 60
    else:
        return int(ts)


def ignore_exceptions(
    retry: int = 0,
    retry_interval: int = 1,
    return_on_error: Any = None,
    logger: Logger | None = None,
    raise_final_exception: bool = False,
) -> Any:
    """
    Decorator to ignore exceptions with support for retry.

    Args:
        retry (int): Number of retry if the underline function throw exception.
        retry_interval (int): Number of seconds to wait between retries.
        return_on_error (Any): Return value if the underline function throw exception.
        logger (Logger): logger to use, if not passed no logs will be displayed.
        raise_final_exception (bool): whether to raise the final exception.


    Returns:
        any: the underline function return value.
    """

    def wrapper(func: Callable) -> Callable:
        @wraps(func)
        def inner(*args: Any, **kwargs: Any) -> Any:
            try:
                return func(*args, **kwargs)
            except Exception as ex:
                for idx in range(0, retry):
                    try:
                        sleep(retry_interval)
                        return func(*args, **kwargs)
                    except Exception as retry_ex:
                        if idx + 1 < retry:
                            continue
                        if raise_final_exception:
                            raise retry_ex
                if logger:
                    logger.error(f"{func.__name__} error: {ex}")
                return return_on_error

        return inner

    return wrapper


def stt(seconds: int) -> str:
    """
    Convert seconds to human readable time string.

    Args:
        seconds (int): seconds to convert

    Returns:
        str: Human readable time string

    Example:
        >>> stt(seconds=3600)
        '1 hour'
        >>> stt(seconds=3600*24)
        '1 day'
        >>> stt((60*60*14+65))
        '14 hours and 1 minute and 5 seconds'
    """
    time_str = ""
    total_seconds = datetime.timedelta(seconds=seconds)
    days = total_seconds.days
    total_time = datetime.datetime.strptime(str(datetime.timedelta(seconds=total_seconds.seconds)), "%H:%M:%S")
    hour, minute, second = total_time.hour, total_time.minute, total_time.second
    if days:
        time_str += f"{days} {'days' if days > 1 else 'day'}"

    if hour:
        if days:
            time_str += " and "
        time_str += f"{hour} {'hours' if hour > 1 else 'hour'}"

    if minute:
        if hour or days:
            time_str += " and "
        time_str += f"{minute} {'minutes' if minute > 1 else 'minute'}"

    if total_time.second:
        if hour or minute:
            time_str += " and "
        time_str += f"{second} {'seconds' if second > 1 else 'second'}"

    return time_str



from __future__ import annotations


class CommandExecFailed(Exception):
    def __init__(self, name: str, err: str | None = None) -> None:
        self.name = name
        self.err = f"Error: {err}" if err else ""

    def __str__(self) -> str:
        return f"Command: {self.name} - exec failed. {self.err}"
